"""Base for VT100 terminals."""

import asyncio
import os
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import Enum, auto
from io import StringIO
from typing import Final, Literal

from ..colors import Color
from ..geometry import Point, Size
from .ansi_escapes import ANSI_ESCAPES
from .events import (
    ColorReportEvent,
    CursorPositionResponseEvent,
    DeviceAttributesReportEvent,
    Event,
    FocusEvent,
    KeyEvent,
    MouseButton,
    MouseEvent,
    MouseEventType,
    PasteEvent,
    UnknownEscapeSequence,
)

CPR_RE: Final = re.compile(r"\x1b\[(\d+);(\d+)R")
COLOR_RE: Final = re.compile(
    r"\x1b\]1([10]);rgb:([0-9a-f]{4})/([0-9a-f]{4})/([0-9a-f]{4})\x1b\\"
)
# \x1b[?61;4;6;7;14;21;22;23;24;28;32;42c
DEVICE_ATTRIBUTES_RE: Final = re.compile(r"\x1b\[\?[0-9;]+c")
MOUSE_SGR_RE: Final = re.compile(r"\x1b\[<(\d+);(\d+);(\d+)(m|M)")
PARAMS_RE: Final = re.compile(r"[0-9;]")
BRACKETED_PASTE_START: Final = "\x1b[200~"
BRACKETED_PASTE_END: Final = "\x1b[201~"
FOCUS_IN: Final = "\x1b[I"
FOCUS_OUT: Final = "\x1b[O"
ESCAPE_TIMEOUT: Final = 0.05
"""Time in seconds before escape buffer is reset."""
DRS_REQUEST_TIMEOUT: Final = 0.1
"""
Time in seconds for the input parser to expect a response to a device status report
request.
"""


class ParserState(Enum):
    """State of VT100 input parser."""

    GROUND = auto()
    """Initial state."""
    ESCAPE = auto()
    """Start of an escape sequence."""
    CSI = auto()
    """Start of a control sequence."""
    OSC = auto()
    """Start of an operating system command sequence."""
    PARAMS = auto()
    """Collecting parameters of a control sequence."""
    PASTE = auto()
    """Collecting paste data."""
    EXECUTE_NEXT = auto()
    """Execute escape buffer after receiving next character."""


class Vt100Terminal(ABC):
    """
    Base for VT100 terminals.

    Attributes
    ----------
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    flush()
        Write buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    def __init__(self):
        self.in_alternate_screen: bool = False
        """Whether the alternate screen buffer is enabled."""
        self._escape_buffer: StringIO | None = None
        """Escape sequence buffer."""
        self._paste_buffer: StringIO | None = None
        """Paste buffer."""
        self._event_buffer: list[Event] = []
        """Events generated during input parsing."""
        self._out_buffer: list[str] = []
        """
        Output buffer.

        Escapes for stdout are collected here before ``flush()`` is called.
        """
        self._state: ParserState = ParserState.GROUND
        """State of VT100 input parser."""
        self._timeout_escape: asyncio.TimerHandle | None = None
        """Timeout handle for executing escape buffer."""
        self._dsr_pending: int = 0
        """Pending device status report requests."""
        self._last_y: int = 0
        """Last mouse y-coordinate."""
        self._last_x: int = 0
        """Laste mouse x-coordinate."""
        self._event_handler: Callable[[list[Event]], None] | None = None
        """Event handler set with ``attach()`` or unset with ``unattach()``."""

    @abstractmethod
    def process_stdin(self) -> None:
        """Read from stdin and feed data into input parser to generate events."""

    @abstractmethod
    def raw_mode(self) -> None:
        """Set terminal to raw mode."""

    @abstractmethod
    def restore_console(self) -> None:
        """Restore console to its original mode."""

    @abstractmethod
    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """

    @abstractmethod
    def unattach(self) -> None:
        """Stop generating events from stdin."""

    def events(self) -> list[Event]:
        """Return a list of input events and reset the event buffer."""
        events = self._event_buffer
        self._event_buffer = []
        return events

    def get_size(self) -> Size:
        """Get terminal size."""
        cols, rows = os.get_terminal_size()
        return Size(rows, cols)

    def _feed(self, data: str) -> None:
        """Generate events from terminal input data."""
        if self._timeout_escape is not None:
            self._timeout_escape.cancel()
            self._timeout_escape = None

        for char in data:
            self._feed1(char)

        if self._state is ParserState.GROUND:
            return

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            pass
        else:
            self._timeout_escape = loop.call_later(ESCAPE_TIMEOUT, self._reset_escape)

    def _feed1(self, char: str) -> None:
        """Feed a single character from terminal input into the parser."""
        if self._state is ParserState.OSC:
            self._escape_buffer.write(char)
            if char == "\\" and self._escape_buffer.getvalue().endswith("\x1b\\"):
                self._execute()
        elif self._state is not ParserState.PASTE and char == "\x1b":
            # Start a new escape (possibly canceling previous escape).
            self._escape_buffer = StringIO()
            self._escape_buffer.write(char)
            self._state = ParserState.ESCAPE
        elif self._state is ParserState.EXECUTE_NEXT:
            self._escape_buffer.write(char)
            self._execute()
        elif self._state is ParserState.PASTE:
            self._paste_buffer.write(char)
            if char == "~":
                paste = self._paste_buffer.getvalue()
                if paste.endswith(BRACKETED_PASTE_END):
                    self._event_buffer.append(PasteEvent(paste[:-6]))
                    self._paste_buffer = None
                    self._state = ParserState.GROUND
        elif self._state is ParserState.GROUND:
            if ord(char) < 0x20 or char == "\x7f" or char == "\x9b":
                self._escape_buffer = StringIO()
                self._escape_buffer.write(char)
                self._execute()
            else:
                self._event_buffer.append(KeyEvent(char))
        elif self._state is ParserState.ESCAPE:
            self._escape_buffer.write(char)
            if char == "[":
                self._state = ParserState.CSI
            elif char == "O":
                self._state = ParserState.EXECUTE_NEXT
            elif char == "]":
                self._state = ParserState.OSC
            else:
                self._execute()
        elif self._state is ParserState.CSI:
            self._escape_buffer.write(char)
            if char == "[":
                self._state = ParserState.EXECUTE_NEXT
            elif char == "<" or char == "?":
                self._state = ParserState.PARAMS
            elif PARAMS_RE.fullmatch(char) is None:
                self._execute()
            else:
                self._state = ParserState.PARAMS
        elif self._state is ParserState.PARAMS:
            self._escape_buffer.write(char)
            if PARAMS_RE.fullmatch(char) is None:
                self._execute()

    def _execute(self) -> None:
        """Produce an event from the escape buffer."""
        self._state = ParserState.GROUND
        escape = self._escape_buffer.getvalue()
        self._escape_buffer = None

        if self._dsr_pending:
            if cpr_match := CPR_RE.fullmatch(escape):
                self._dsr_pending -= 1
                y, x = cpr_match.groups()
                self._event_buffer.append(
                    CursorPositionResponseEvent(Point(int(y) - 1, int(x) - 1))
                )
                return

            if color_match := COLOR_RE.fullmatch(escape):
                self._dsr_pending -= 1
                kind, r, g, b = color_match.groups()
                self._event_buffer.append(
                    ColorReportEvent(
                        kind="fg" if kind == "0" else "bg",
                        color=Color.from_hex(f"{r[:2]}{g[:2]}{b[:2]}"),
                    )
                )
                return

            if device_attributes_match := DEVICE_ATTRIBUTES_RE.fullmatch(escape):
                self._dsr_pending -= 1
                device_attributes = device_attributes_match.group()[3:-1].split(";")
                self._event_buffer.append(
                    DeviceAttributesReportEvent(frozenset(map(int, device_attributes)))
                )
                return

        if escape == BRACKETED_PASTE_START:
            self._state = ParserState.PASTE
            self._paste_buffer = StringIO(newline=None)
        elif escape == FOCUS_IN:
            self._event_buffer.append(FocusEvent("in"))
        elif escape == FOCUS_OUT:
            self._event_buffer.append(FocusEvent("out"))
        elif sgr_match := MOUSE_SGR_RE.fullmatch(escape):
            info = int(sgr_match[1])
            y = int(sgr_match[3]) - 1
            x = int(sgr_match[2]) - 1
            dy = y - self._last_y
            dx = x - self._last_x
            self._last_y = y
            self._last_x = x
            state = sgr_match[4]
            button: MouseButton = ["left", "middle", "right", "no_button"][info % 4]
            event_type: MouseEventType

            if info & 64:
                event_type = "scroll_down" if info & 1 else "scroll_up"
                button = "no_button"
            elif info & 32:
                event_type = "mouse_move"
            elif state == "m":
                event_type = "mouse_up"
            else:
                event_type = "mouse_move" if button == "no_button" else "mouse_down"

            shift = bool(info & 4)
            alt = bool(info & 8)
            ctrl = bool(info & 16)

            self._event_buffer.append(
                MouseEvent(Point(y, x), button, event_type, alt, ctrl, shift, dy, dx)
            )
        elif escape in ANSI_ESCAPES:
            self._event_buffer.append(KeyEvent(*ANSI_ESCAPES[escape]))
        elif len(escape) == 2 and 32 <= ord(escape[1]) <= 126:
            self._event_buffer.append(KeyEvent(escape[1], alt=True))
        else:
            self._event_buffer.append(UnknownEscapeSequence(escape))

    def _reset_escape(self) -> None:
        """Execute escape buffer after a timeout period."""
        if self._state is ParserState.PASTE:
            paste = self._paste_buffer.getvalue()
            self._paste_buffer = None
            self._state = ParserState.GROUND

            # Timed out during a paste. Check if there's a partial escape to remove
            # (maybe BRACKETED_PASTE_END was cutoff).
            partial_escape_index = paste.find("\x1b")
            if partial_escape_index != -1:
                ending = paste[partial_escape_index:]
                if BRACKETED_PASTE_END[: len(ending)] == ending:
                    paste = paste[:partial_escape_index]
            self._event_buffer.append(PasteEvent(paste))
            self._paste_buffer = None
        else:
            self._execute()

        if self._event_handler is not None:
            self._event_handler(self.events())

    def flush(self) -> None:
        """Write buffer to output stream and flush."""
        if not self._out_buffer:
            return

        data = "".join(self._out_buffer).encode(errors="replace")
        self._out_buffer.clear()
        sys.__stdout__.buffer.write(data)
        sys.__stdout__.flush()

    def set_title(self, title: str) -> None:
        """Set terminal title."""
        self._out_buffer.append(f"\x1b]2;{title}\x07")

    def enter_alternate_screen(self) -> None:
        """Enter alternate screen buffer."""
        self._out_buffer.append("\x1b[?1049h\x1b[H")
        self.in_alternate_screen = True

    def exit_alternate_screen(self) -> None:
        """Exit alternate screen buffer."""
        self._out_buffer.append("\x1b[?1049l")
        self.in_alternate_screen = False

    def enable_mouse_support(self) -> None:
        """Enable mouse support in terminal."""
        self._out_buffer.append(
            "\x1b[?1000h"  # SET_VT200_MOUSE
            "\x1b[?1003h"  # SET_ANY_EVENT_MOUSE
            "\x1b[?1006h"  # SET_SGR_EXT_MODE_MOUSE
        )

    def disable_mouse_support(self) -> None:
        """Disable mouse support in terminal."""
        self._out_buffer.append(
            "\x1b[?1000l"  # SET_VT200_MOUSE
            "\x1b[?1003l"  # SET_ANY_EVENT_MOUSE
            "\x1b[?1006l"  # SET_SGR_EXT_MODE_MOUSE
        )

    def reset_attributes(self) -> None:
        """Reset character attributes."""
        self._out_buffer.append("\x1b[0m")

    def enable_bracketed_paste(self) -> None:
        """Enable bracketed paste in terminal."""
        self._out_buffer.append("\x1b[?2004h")

    def disable_bracketed_paste(self) -> None:
        """Disable bracketed paste in terminal."""
        self._out_buffer.append("\x1b[?2004l")

    def show_cursor(self) -> None:
        """Show cursor in terminal."""
        self._out_buffer.append("\x1b[?25h")

    def hide_cursor(self) -> None:
        """Hide cursor in terminal."""
        self._out_buffer.append("\x1b[?25l")

    def enable_reporting_focus(self) -> None:
        """Enable reporting terminal focus."""
        self._out_buffer.append("\x1b[?1004h")

    def disable_reporting_focus(self) -> None:
        """Disable reporting terminal focus."""
        self._out_buffer.append("\x1b[?1004l")

    def _dsr_request(self, escape: str) -> None:
        """Make a device status report request and schedule its cancellation."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            return

        self._dsr_pending += 1
        loop.call_later(DRS_REQUEST_TIMEOUT, self._timeout_dsr)
        self._out_buffer.append(escape)
        self.flush()

    def _timeout_dsr(self) -> None:
        if self._dsr_pending:
            self._dsr_pending -= 1

    def request_cursor_position_report(self) -> None:
        """Report current cursor position."""
        self._dsr_request("\x1b[6n")

    def request_foreground_color(self) -> None:
        """Report terminal foreground color."""
        self._dsr_request("\x1b]10;?\x1b\\")

    def request_background_color(self) -> None:
        """Report terminal background color."""
        self._dsr_request("\x1b]11;?\x1b\\")

    def request_device_attributes(self) -> None:
        """Report device attributes."""
        self._dsr_request("\x1b[c")

    def expect_dsr(self) -> bool:
        """Return whether a device status report is expected."""
        return bool(self._dsr_pending)

    def move_cursor(self, pos: Point) -> None:
        """
        Move cursor to ``pos``.

        Parameters
        ----------
        pos : Point | None, default: None
            Cursor's new position.
        """
        y, x = pos
        self._out_buffer.append(f"\x1b[{y + 1};{x + 1}H")

    def erase_in_display(self, n: Literal[0, 1, 2, 3] = 0) -> None:
        """
        Clear part of the screen.

        Parameters
        ----------
        n : int, default: 0
            Determines which part of the screen to clear. If n is ``0``, clear from
            cursor to end of the screen. If n is ``1``, clear from cursor to beginning
            of the screen. If n is ``2``, clear entire screen. If n is ``3``, clear
            entire screen and delete all lines in scrollback buffer.
        """
        self._out_buffer.append(f"\x1b[{n}J")



"""An image painted with box unicode characters."""

from pathlib import Path

import cv2
import numpy as np

from ..text_tools import binary_to_box
from ..texture_tools import Interpolation, resize_texture
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .text import Text

__all__ = ["BoxImage", "Interpolation", "Point", "Size"]


class BoxImage(Gadget):
    r"""
    An image painted with box unicode characters.

    Parameters
    ----------
    path : pathlib.Path
        Path to image.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    path : pathlib.Path
        Path to image.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._image = Text(is_transparent=is_transparent)
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.add_gadget(self._image)
        self.alpha = alpha
        self.interpolation = interpolation
        self.path = path

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._image.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._image.alpha = alpha

    @property
    def interpolation(self) -> Interpolation:
        """Interpolation used when gadget is resized."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, interpolation: Interpolation):
        if interpolation not in Interpolation.__args__:
            raise TypeError(f"{interpolation} is not a valid interpolation type.")
        self._interpolation = interpolation

    @property
    def path(self) -> Path | None:
        """Path to image."""
        return self._path

    @path.setter
    def path(self, path: Path | None):
        self._path: Path | None = path

        if path is None:
            self._otexture = np.zeros((1, 1, 3), dtype=np.uint8)
        else:
            self._otexture = cv2.imread(str(path.absolute()), cv2.IMREAD_COLOR)
        self._load_texture()

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._image.is_transparent = self.is_transparent

    def on_size(self):
        """Remake canvas."""
        self._image.size = self.size
        self._load_texture()

    def _load_texture(self):
        h, w = self.size
        if h == 0 or w == 0:
            return

        canvas = self._image.canvas
        img_bgr = resize_texture(self._otexture, (2 * h, 2 * w), self.interpolation)
        img_rgb = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2RGB)
        img_hls = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2HLS)

        rgb_sectioned = np.swapaxes(img_rgb.reshape(h, 2, w, 2, 3), 1, 2)
        hls_sectioned = np.swapaxes(img_hls.reshape(h, 2, w, 2, 3), 1, 2)

        # First, find the average lightness of each 2x2 section of the image
        # (`average_lightness`). Boxes are placed where the lightness is greater than
        # `average_lightness`. The background color will be the average of the colors
        # darker than `average_lightness`. The foreground color will be the average of
        # the colors lighter than `average_lightness`.

        lightness = hls_sectioned[..., 1]
        average_lightness = np.average(lightness, axis=(2, 3))
        where_boxes = lightness > average_lightness[..., None, None]

        canvas["char"] = binary_to_box(where_boxes)

        nboxes = where_boxes.sum(axis=(2, 3))
        nboxes_neg = 4 - nboxes
        nboxes[nboxes == 0] = 1
        nboxes_neg[nboxes_neg == 0] = 1

        foreground = rgb_sectioned.copy()
        foreground[~where_boxes] = 0
        canvas["fg_color"] = foreground.sum(axis=(2, 3)) / nboxes[..., None]

        background = rgb_sectioned.copy()
        background[where_boxes] = 0
        canvas["bg_color"] = background.sum(axis=(2, 3)) / nboxes_neg[..., None]



"""A text gadget."""

from typing import Literal

import numpy as np
from numpy.typing import NDArray
from pygments.lexer import Lexer
from pygments.lexers import guess_lexer
from pygments.style import Style

from ..colors import Color, Neptune
from ..geometry import rect_slice
from ..text_tools import (
    _parse_batgrl_md,
    _text_to_cells,
    _write_lines_to_canvas,
    add_text,
    cell_sans,
    char_width,
    coerce_cell,
    str_width,
)
from ..texture_tools import _composite
from .gadget import (
    Cell,
    Gadget,
    Point,
    PosHint,
    Size,
    SizeHint,
    bindable,
    clamp,
    new_cell,
)

__all__ = [
    "Text",
    "Border",
    "Cell",
    "Point",
    "Size",
    "add_text",
    "char_width",
    "str_width",
]

Border = Literal[
    "light",
    "heavy",
    "double",
    "curved",
    "ascii",
    "outer",
    "inner",
    "thick",
    "dashed",
    "dashed_2",
    "dashed_3",
    "heavy_dashed",
    "heavy_dashed_2",
    "heavy_dashed_3",
    "near",
    "mcgugan_tall",
    "mcgugan_wide",
]
"""Border styles for :meth:`batgrl.text_gadget.Text.add_border`."""

_BORDERS = {
    "light": "┌┐││──└┘",
    "heavy": "┏┓┃┃━━┗┛",
    "double": "╔╗║║══╚╝",
    "curved": "╭╮││──╰╯",
    "ascii": "++||--++",
    "outer": "▛▜▌▐▀▄▙▟",
    "inner": "▗▖▐▌▄▀▝▘",
    "thick": "████▀▄██",
    "dashed": "┌┐╎╎╌╌└┘",
    "dashed_2": "┌┐┆┆┄┄└┘",
    "dashed_3": "┌┐┊┊┈┈└┘",
    "heavy_dashed": "┏┓╏╏╍╍┗┛",
    "heavy_dashed_2": "┏┓┇┇┅┅┗┛",
    "heavy_dashed_3": "┏┓┋┋┉┉┗┛",
    "near": "  ▕▏▁▔  ",
    "mcgugan_tall": "▕▏▕▏▔▁▕▏",
    "mcgugan_wide": "▁▁▏▕▁▔▔▔",
}
"""Border characters for :meth:`batgrl.text_gadget.Text.add_border`."""


class Text(Gadget):
    r"""
    A text gadget. Displays arbitrary text data.

    Parameters
    ----------
    default_cell : NDArray[Cell] | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 0.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    canvas : NDArray[Cell]
        The array of characters for the gadget.
    default_cell : NDArray[Cell]
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        default_cell: NDArray[Cell] | str = " ",
        alpha: float = 0.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        size = self.size
        self.default_cell = default_cell
        self.canvas = np.full(size, self.default_cell)
        self.alpha = alpha

    @property
    def default_cell(self) -> NDArray[Cell]:
        """Default character for text canvas."""
        return self._default_cell

    @default_cell.setter
    def default_cell(self, cell: NDArray[Cell] | str):
        self._default_cell = coerce_cell(cell, new_cell())

    @property
    def default_fg_color(self) -> Color:
        """Foreground color of default character."""
        return Color(*self._default_cell["fg_color"])

    @default_fg_color.setter
    def default_fg_color(self, default_fg_color: Color):
        self._default_cell["fg_color"] = default_fg_color

    @property
    def default_bg_color(self) -> Color:
        """Background color of default character."""
        return Color(*self._default_cell["bg_color"])

    @default_bg_color.setter
    def default_bg_color(self, default_bg_color: Color):
        self._default_cell["bg_color"] = default_bg_color

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    def on_size(self):
        """Resize canvas preserving as much content as possible."""
        old_canvas = self.canvas
        old_h, old_w = old_canvas.shape

        h, w = self._size

        copy_h = min(old_h, h)
        copy_w = min(old_w, w)

        self.canvas = np.full((h, w), self.default_cell)
        self.canvas[:copy_h, :copy_w] = old_canvas[:copy_h, :copy_w]

    def add_border(
        self,
        style: Border = "light",
        bold: bool = False,
        fg_color: Color | None = None,
        bg_color: Color | None = None,
    ):
        """
        Add a text border (gadget height and width must be at least 2).

        Parameters
        ----------
        style : Border, default: "light"
            Style of border. Default style uses light box-drawing characters.
        bold : bool, default: False
            Whether the border is bold.
        fg_color : Color | None, default: None
            Foreground color of border if not None.
        bg_color : Color | None, default: None
            Background color of border if not None.
        """
        if self.height < 2 or self.width < 2:
            return

        slices = [
            np.s_[0, 0],
            np.s_[0, -1],
            np.s_[1:-1, 0],
            np.s_[1:-1, -1],
            np.s_[0, 1:-1],
            np.s_[-1, 1:-1],
            np.s_[-1, 0],
            np.s_[-1, -1],
        ]
        view = cell_sans("fg_color", "bg_color")
        cells = self.canvas[view]
        for s, border in zip(slices, _BORDERS[style]):
            cells[s] = new_cell(char=border, bold=bold)[view]

        if fg_color is not None:
            self.canvas["fg_color"][[0, -1]] = fg_color
            self.canvas["fg_color"][:, [0, -1]] = fg_color
        if bg_color is not None:
            self.canvas["bg_color"][[0, -1]] = bg_color
            self.canvas["bg_color"][:, [0, -1]] = bg_color

    def add_syntax_highlighting(
        self, lexer: Lexer | None = None, style: Style = Neptune
    ):
        """
        Add syntax highlighting to current text in canvas.

        Parameters
        ----------
        lexer : pygments.lexer.Lexer | None, default: None
            Lexer for text. If not given, the lexer is guessed.
        style : pygments.style.Style, default: Neptune
            A pygments style to use for syntax highlighting.
        """
        text = "\n".join("".join(line).rstrip() for line in self.canvas["char"])
        if lexer is None:
            lexer = guess_lexer(text)

        self.canvas["fg_color"] = 0
        self.canvas["bg_color"] = Color.from_hex(style.background_color)
        y = x = 0
        for ttype, value in lexer.get_tokens(text):
            lines = value.split("\n")
            token_style = style.style_for_token(ttype)
            for i, line in enumerate(lines):
                if i > 0:
                    y += 1
                    x = 0

                if len(line) == 0:
                    continue

                end = x + str_width(line)
                if token_style["color"]:
                    self.canvas[y, x:end]["fg_color"] = Color.from_hex(
                        token_style["color"]
                    )
                if token_style["bgcolor"]:
                    self.canvas[y, x:end]["bg_color"] = Color.from_hex(
                        token_style["bgcolor"]
                    )
                self.canvas[y, x:end]["bold"] = token_style["bold"]
                self.canvas[y, x:end]["italic"] = token_style["italic"]
                self.canvas[y, x:end]["underline"] = token_style["underline"]
                x = end

    def add_str(
        self,
        str: str,
        *,
        pos: Point = Point(0, 0),
        fg_color: Color | None = None,
        bg_color: Color | None = None,
        markdown: bool = False,
        truncate_str: bool = False,
    ):
        """
        Add a single line of text to the canvas at position `pos`.

        If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
        - italic: `*this is italic text*`
        - bold: `**this is bold text**`
        - strikethrough: `~~this is strikethrough text~~`
        - underlined: `__this is underlined text__`
        - overlined: `^^this is overlined text^^`

        Parameters
        ----------
        str : str
            A single line of text to add to canvas.
        pos : Point, default: Point(0, 0)
            Position of first character of string. Negative coordinates position
            from the right or bottom of canvas (like negative indices).
        fg_color : Color | None, default: None
            Foreground color of text.
        bg_color : Color | None, default: None
            Background color of text.
        markdown : bool, default: False
            Whether to parse text for batgrl markdown.
        truncate_str : bool, default: False
            If false, an `IndexError` is raised if the text would not fit on canvas.

        See Also
        --------
        text_tools.add_text : Add multiple lines of text to a view of a canvas.
        """
        y, x = pos
        add_text(
            self.canvas[y, x:],
            str,
            fg_color=fg_color,
            bg_color=bg_color,
            markdown=markdown,
            truncate_text=truncate_str,
        )

    def set_text(
        self,
        text: str,
        *,
        fg_color: Color | None = None,
        bg_color: Color | None = None,
        markdown: bool = False,
    ):
        """
        Resize gadget to fit text, erase canvas, then fill canvas with text.

        If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
        - italic: `*this is italic text*`
        - bold: `**this is bold text**`
        - strikethrough: `~~this is strikethrough text~~`
        - underlined: `__this is underlined text__`
        - overlined: `^^this is overlined text^^`

        Parameters
        ----------
        text : str
            Text to add to canvas.
        fg_color : Color | None, default: None
            Foreground color of text.
        bg_color : Color | None, default: None
            Background color of text.
        markdown : bool, default: False
            Whether to parse text for batgrl markdown.

        See Also
        --------
        text_tools.add_text : Add multiple lines of text to a view of a canvas.
        """
        self.size, lines = _parse_batgrl_md(text) if markdown else _text_to_cells(text)
        self.clear()
        _write_lines_to_canvas(lines, self.canvas, fg_color, bg_color)

    def clear(self):
        """Fill canvas with default cell."""
        self.canvas[:] = self.default_cell

    def shift(self, n: int = 1):
        """
        Shift content in canvas up (or down in case of negative `n`).

        Rows at the bottom (or top) will be filled with the default cell.
        """
        if n > 0:
            self.canvas[:-n] = self.canvas[n:]
            self.canvas[-n:] = self.default_cell
        elif n < 0:
            self.canvas[-n:] = self.canvas[:n]
            self.canvas[:-n] = self.default_cell

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        sans_bg = canvas[cell_sans("bg_color")]
        foreground = canvas["fg_color"]
        background = canvas["bg_color"]
        text_chars = self.canvas["char"]
        text_sans_bg = self.canvas[cell_sans("bg_color")]
        text_bg = self.canvas["bg_color"]
        root_pos = self.root._pos
        abs_pos = self.absolute_pos
        alpha = self.alpha
        for pos, size in self._region.rects():
            dst = rect_slice(pos - root_pos, size)
            src = rect_slice(pos - abs_pos, size)
            if self.is_transparent:
                visible = np.isin(text_chars[src], (" ", "⠀"), invert=True)
                invisible = ~visible
                sans_bg[dst][visible] = text_sans_bg[src][visible]
                fg = foreground[dst][invisible]  # Not a view.
                _composite(fg, text_bg[src][invisible], 255, alpha)
                foreground[dst][invisible] = fg
                _composite(background[dst], text_bg[src], 255, alpha)
            else:
                canvas[dst] = self.canvas[src]



"""A gadget that re-colors cells beneath it."""

from numpy.typing import NDArray

from ..colors import BLACK, WHITE, Color
from ..geometry import rect_slice
from .gadget import Cell, Gadget


class Cursor(Gadget):
    """A gadget that re-colors cells beneath it."""

    def __init__(self, fg_color: Color = WHITE, bg_color: Color = BLACK, **kwargs):
        super().__init__(size=(1, 1), is_transparent=True, is_enabled=False, **kwargs)
        self.fg_color: Color = fg_color
        """Foreground color of gadget."""
        self.bg_color: Color = bg_color
        """Background color of gadget."""

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        root_pos = self.root._pos
        abs_pos = self.parent.absolute_pos
        for pos, size in self._region.rects():
            dst = rect_slice(pos - root_pos, size)
            src = rect_slice(pos - abs_pos, size)
            canvas[dst]["fg_color"] = self.fg_color
            canvas[dst]["bg_color"] = self.bg_color
            canvas[dst]["char"] = self.parent.canvas[src]["char"]



"""
A text particle field.

A particle field specializes in handling many single "pixel" children.
"""

from typing import Any

import numpy as np
from numpy.typing import NDArray

from ..geometry import clamp, rect_slice
from ..text_tools import cell_sans
from ..texture_tools import _composite
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, new_cell

__all__ = ["TextParticleField", "particle_data_from_canvas", "Point", "Size"]


class TextParticleField(Gadget):
    r"""
    A text particle field.

    A particle field specializes in rendering many single "pixel" children. This is more
    efficient than rendering many 1x1 gadgets.

    Parameters
    ----------
    particle_positions : NDArray[np.int32] | None, default: None
        An array of particle positions with shape `(N, 2)`.
    particle_cells : NDArray[Cell] | None, default: None
        An array of Cells of particles with shape `(N,)`.
    particle_properties : dict[str, Any] | None, default: None
        Additional particle properties.
    alpha : float, default: 0.0
        Transparency of particles.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    nparticles : int
        Number of particles in particle field.
    particle_positions : NDArray[np.int32]
        An array of particle positions with shape `(N, 2)`.
    particle_cells : NDArray[Cell]
        An array of Cells of particles with shape `(N,)`.
    particle_properties : dict[str, Any]
        Additional particle properties.
    alpha : float
        Transparency of particles.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    particles_from_cells(cells)
        Return positions and cells of non-whitespace characters of a Cell array.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        particle_positions: NDArray[np.int32] | None = None,
        particle_cells: NDArray[Cell] | None = None,
        particle_properties: dict[str, Any] = None,
        alpha: float = 0.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        if particle_positions is None:
            self.particle_positions = np.zeros((0, 2), dtype=int)
        else:
            self.particle_positions = np.asarray(particle_positions, dtype=int)

        if particle_cells is None:
            self.particle_cells = np.full(len(self.particle_positions), new_cell())
        else:
            self.particle_cells = np.asarray(particle_cells, dtype=Cell)

        if particle_properties is None:
            self.particle_properties = {}
        else:
            self.particle_properties = particle_properties

        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of particles."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._alpha = clamp(alpha, 0.0, 1.0)

    @property
    def nparticles(self) -> int:
        """Number of particles in particle field."""
        return len(self.particle_positions)

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        chars = canvas[cell_sans("bg_color")]
        bg_color = canvas["bg_color"]
        root_pos = self.root._pos
        abs_pos = self.absolute_pos
        ppos = self.particle_positions
        pchars = self.particle_cells[cell_sans("bg_color")]
        pbg_color = self.particle_cells["bg_color"]
        for pos, size in self._region.rects():
            abs_ppos = ppos - (pos - abs_pos)
            inbounds = (((0, 0) <= abs_ppos) & (abs_ppos < size)).all(axis=1)

            if self.is_transparent:
                not_whitespace = np.isin(pchars["char"], (" ", "⠀"), invert=True)
                where_inbounds = np.nonzero(inbounds & not_whitespace)
            else:
                where_inbounds = np.nonzero(inbounds)
            painted = pbg_color[where_inbounds]

            ys, xs = abs_ppos[where_inbounds].T
            dst = rect_slice(pos - root_pos, size)
            if self.is_transparent:
                background = bg_color[dst][ys, xs]
                _composite(background, painted, 255, self.alpha)
                bg_color[dst][ys, xs] = background
            else:
                bg_color[dst][ys, xs] = painted

            chars[dst][ys, xs] = pchars[where_inbounds]


def particle_data_from_canvas(
    canvas: NDArray[Cell],
) -> tuple[NDArray[np.int32], NDArray[Cell]]:
    """
    Return positions and cells of non-whitespace characters of a Cell array.

    Parameters
    ----------
    canvas : NDArray[Cell]
        A Cell numpy array (such as a :class:`Text` gadget's canvas).

    Returns
    -------
    tuple[NDArray[np.int32], NDArray[Cell]]
        Position and cells of non-whitespace characters of the canvas.
    """
    positions = np.argwhere(np.isin(canvas["char"], (" ", "⠀"), invert=True))
    pys, pxs = positions.T
    return positions, canvas[pys, pxs]



"""A tabbed gadget."""

import asyncio

from ..colors import lerp_colors
from .behaviors.themable import Themable
from .behaviors.toggle_button_behavior import ToggleButtonBehavior
from .gadget import Gadget, Point, PosHint, Size, SizeHint, new_cell
from .pane import Pane
from .text import Text

__all__ = ["Tabs", "Point", "Size"]

# TODO: Movable tabs?
TAB_SPACING = 3


class _Tab(Themable, ToggleButtonBehavior, Text):
    def __init__(self, title, content, **kwargs):
        super().__init__(**kwargs)
        self.title = title
        self.content = content
        self.set_text(title)

    def on_toggle(self):
        if self.parent is None:
            return
        if self.toggle_state == "on":
            self.content.is_enabled = True

            tabbed: Tabs = self.parent.parent

            tabbed._history.remove(self.title)
            tabbed._history.append(self.title)

            tabbed._active_tab = self.title

            if tabbed._underline_task is not None:
                tabbed._underline_task.cancel()
            tabbed._underline_task = asyncio.create_task(
                tabbed._tab_underline.tween(
                    duration=0.5,
                    easing="out_bounce",
                    x=self.x - 1,
                    width=self.width + 2,
                )
            )
        else:
            self.content.is_enabled = False

    def _update(self):
        if self.toggle_state == "on":
            color_pair = self.color_theme.titlebar_normal
            bold = True
        elif self.button_state == "hover":
            color_pair = self.hover_color_pair
            bold = False
        else:
            color_pair = self.color_theme.titlebar_inactive
            bold = False
        self.canvas[["fg_color", "bg_color"]] = color_pair
        self.canvas["bold"] = bold

    update_hover = _update
    update_normal = _update
    update_on = _update
    update_off = _update

    def update_theme(self):
        normal = self.color_theme.titlebar_normal
        inactive = self.color_theme.titlebar_inactive
        self.hover_color_pair = (
            lerp_colors(normal.fg, inactive.fg, 0.5),
            lerp_colors(normal.bg, inactive.bg, 0.5),
        )
        self._update()


class Tabs(Themable, Gadget):
    r"""
    A tabbed gadget.

    Parameters
    ----------
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    add_tab(title, content)
        Add a new tab.
    remove_tab(title)
        Remove a tab.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.tabs: dict[str, Gadget] = {}

        h, w = self.size
        self.tab_bar = Pane(size_hint={"width_hint": 1.0}, size=(1, w))
        self.separator = Text(size_hint={"width_hint": 1.0}, size=(1, w), pos=(1, 0))

        def _update_sep():
            self.separator.canvas["char"] = "━"

        self.bind("size", _update_sep)

        self.tab_window = Pane(
            size=(h - 2, w),
            pos=(2, 0),
            size_hint={"height_hint": 1.0, "width_hint": 1.0, "height_offset": -2},
        )

        title_fg, title_bg = self.color_theme.titlebar_normal
        tab_style = {"bold": True, "fg_color": title_fg, "bg_color": title_bg}
        self._tab_underline = Text(
            size=(1, 1),
            pos=(1, 0),
            is_enabled=False,
            default_cell=new_cell(char="━", **tab_style),
        )
        tab_underline_left = Text(
            size=(1, 1),
            default_cell=new_cell(char="╺", **tab_style),
        )
        tab_underline_right = Text(
            size=(1, 1),
            pos_hint={"x_hint": 1.0, "anchor": "right"},
            default_cell=new_cell(char="╸", **tab_style),
        )
        self._tab_underline.add_gadgets(tab_underline_left, tab_underline_right)

        self._active_tab = None
        self._history = []  # Used to select the last viewed tab when a tab is removed.
        self._underline_task = None

        self.add_gadgets(
            self.tab_bar, self.separator, self._tab_underline, self.tab_window
        )

    def _reposition_tabs(self):
        x = 1
        for child in self.tab_bar.children:
            child.x = x
            x += child.width + TAB_SPACING

    def update_theme(self):
        """Paint the gadget with current theme."""
        title_inactive = self.color_theme.titlebar_inactive
        self.tab_bar.bg_color = title_inactive.bg
        self.tab_window.bg_color = title_inactive.bg
        self.separator.default_cell[["fg_color", "bg_color"]] = title_inactive
        self.separator.canvas[["fg_color", "bg_color"]] = title_inactive

        title_active = self.color_theme.titlebar_normal
        self._tab_underline.default_cell[["fg_color", "bg_color"]] = title_active
        self._tab_underline.canvas[["fg_color", "bg_color"]] = title_active

    def add_tab(self, title: str, content: Gadget):
        """
        Add a new tab. Tab titles are unique.

        Parameters
        ----------
        title : str
            Title of tab.
        content : Gadget
            Content of the tab.
        """
        if title in self.tabs:
            raise ValueError("Tab already exists.")

        tab = _Tab(title, content, group=id(self))
        self.tabs[title] = tab

        self.tab_bar.add_gadget(tab)
        self._reposition_tabs()

        self.tab_window.add_gadget(content)
        self._history.append(title)
        self.tabs[title].toggle_state = "on"
        self._tab_underline.is_enabled = True

    def remove_tab(self, title: str):
        """
        Remove a tab.

        Parameters
        ----------
        title : str
            Title of the tab to remove.
        """
        if title not in self.tabs:
            raise ValueError("Tab doesn't exist.")

        tab = self.tabs.pop(title)
        self.tab_window.remove_gadget(tab.content)
        self.tab_bar.remove_gadget(tab)
        self._history.remove(title)
        self._reposition_tabs()

        if self._active_tab is title:
            if self._history:
                self._active_tab = self._history[-1]
                self.tabs[self._active_tab].toggle_state = "on"
            else:
                self._active_tab = None
                self._tab_underline.is_enabled = False
        else:
            self._tab_underline.x = self.tabs[self._active_tab].x - 1



"""A parallax gadget."""

from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Self

import numpy as np
from numpy.typing import NDArray

from .gadget import (
    Cell,
    Gadget,
    Point,
    PosHint,
    Region,
    Size,
    SizeHint,
    bindable,
    clamp,
)
from .image import Image, Interpolation

__all__ = ["Parallax", "Interpolation", "Point", "Size"]


def _check_layer_speeds(
    layers: Sequence[Image], speeds: Sequence[float] | None
) -> Sequence[float]:
    """
    Raise `ValueError` if `layers` and `speeds` are incompatible, else return a sequence
    of layer speeds.
    """
    nlayers = len(layers)
    if speeds is None:
        return [1 / (nlayers - i) for i in range(nlayers)]

    if len(speeds) != nlayers:
        raise ValueError("number of layers doesn't match number of layer speeds")

    return speeds


class Parallax(Gadget):
    r"""
    A parallax gadget.

    Parameters
    ----------
    path : Path | None, default: None
        Path to directory of images for layers of the parallax (loaded
        in lexographical order of filenames) layered from background to foreground.
    speeds : Sequence[float] | None, default: None
        The scrolling speed of each layer. Default speeds are `1/(N - i)` where `N` is
        the number of layers and `i` is the index of a layer.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    offset : tuple[float, float]
        Vertical and horizontal offset of first layer of the parallax.
    layers : list[Image]
        Layers of the parallax.
    speeds : Sequence[float]
        The scrolling speed of each layer.
    vertical_offset : float
        Vertical offset of first layer of the parallax.
    horizontal_offset : float
        Horizontal offset of first layer of the parallax.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    from_textures(textures, ...)
        Create a :class:`Parallax` from an iterable of uint8 RGBA numpy array.
    from_images(images, ...)
        Create a :class:`Parallax` from an iterable of :class:`Image`.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path | None = None,
        speeds: Sequence[float] | None = None,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        is_transparent: bool = True,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.layers: list[Image]
        """Layers of the parallax."""

        if path is None:
            self.layers = []
        else:
            paths = sorted(path.iterdir(), key=lambda file: file.name)
            self.layers = [Image(path=path, is_transparent=True) for path in paths]
            for layer in self.layers:
                layer.parent = self

        super().__init__(
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.speeds = _check_layer_speeds(self.layers, speeds)
        self.alpha = alpha
        self.interpolation = interpolation
        self._vertical_offset = self._horizontal_offset = 0.0
        self.on_size()

    @property
    def _region(self) -> Region:
        """The visible portion of the gadget on the screen."""
        return self._region_value

    @_region.setter
    def _region(self, region: Region):
        self._region_value = region
        for layer in self.layers:
            layer._region = region

    def on_size(self):
        """Resize parallax layers."""
        for layer in self.layers:
            layer.size = self._size
        self._otextures = [layer.texture for layer in self.layers]

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)
        for layer in self.layers:
            layer.alpha = alpha

    @property
    def interpolation(self) -> Interpolation:
        """Interpolation used when gadget is resized."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, interpolation: Interpolation):
        if interpolation not in {"nearest", "linear", "cubic", "area", "lanczos"}:
            raise ValueError(f"{interpolation} is not a valid interpolation type.")
        for layer in self.layers:
            layer.interpolation = interpolation

    @property
    def vertical_offset(self) -> float:
        """Vertical offset of first layer of the parallax."""
        return self._vertical_offset

    @vertical_offset.setter
    def vertical_offset(self, offset: float):
        self._vertical_offset = offset
        self._adjust()

    @property
    def horizontal_offset(self) -> float:
        """Horizontal offset of first layer of the parallax."""
        return self._horizontal_offset

    @horizontal_offset.setter
    def horizontal_offset(self, offset: float):
        self._horizontal_offset = offset
        self._adjust()

    @property
    def offset(self) -> tuple[float, float]:
        """
        Vertical and horizontal offset of first layer of the parallax.

        Other layers will be adjusted automatically when offset is set.
        """
        return self._vertical_offset, self._horizontal_offset

    @offset.setter
    def offset(self, offset: tuple[float, float]):
        self._vertical_offset, self._horizontal_offset = offset
        self._adjust()

    def _adjust(self):
        for speed, texture, layer in zip(
            self.speeds,
            self._otextures,
            self.layers,
        ):
            rolls = (
                -round(speed * self._vertical_offset),
                -round(speed * self._horizontal_offset),
            )
            layer.texture = np.roll(texture, rolls, axis=(0, 1))

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        if self.layers:
            for layer in self.layers:
                layer._render(canvas)
        else:
            super()._render(canvas)

    @classmethod
    def from_textures(
        cls,
        textures: Iterable[NDArray[np.uint8]],
        *,
        speeds: Sequence[float] | None = None,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        is_transparent: bool = True,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Parallax` from an iterable of uint8 RGBA numpy array.

        Parameters
        ----------
        textures : Iterable[NDArray[np.uint8]]
            An iterable of RGBA textures that will be the layers of the parallax.
        speeds : Sequence[float] | None, default: None
            The scrolling speed of each layer. Default speeds are `1/(N - i)` where `N`
            is the number of layers and `i` is the index of a layer.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Parallax
            A new parallax gadget.
        """
        parallax = cls(
            alpha=alpha,
            interpolation=interpolation,
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        parallax.layers = [
            Image.from_texture(
                texture,
                size=parallax.size,
                alpha=parallax.alpha,
                interpolation=parallax.interpolation,
            )
            for texture in textures
        ]
        for layer in parallax.layers:
            layer.parent = parallax
        parallax.speeds = _check_layer_speeds(parallax.layers, speeds)
        return parallax

    @classmethod
    def from_images(
        cls,
        images: Iterable[Image],
        *,
        speeds: Sequence[float] | None = None,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        is_transparent: bool = True,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Parallax` from an iterable of :class:`Image`.

        Parameters
        ----------
        textures : Iterable[Image]
            An iterable of images that will be the layers of the parallax.
        speeds : Sequence[float] | None, default: None
            The scrolling speed of each layer. Default speeds are `1/(N - i)` where `N`
            is the number of layers and `i` is the index of a layer.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Parallax
            A new parallax gadget.
        """
        parallax = cls(
            alpha=alpha,
            interpolation=interpolation,
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        parallax.layers = list(images)
        for image in parallax.layers:
            image.interpolation = parallax.interpolation
            image.size = parallax.size
            image.alpha = parallax.alpha
            image.parent = parallax
        parallax.speeds = _check_layer_speeds(parallax.layers, speeds)
        return parallax



"""A raycaster gadget."""

import numpy as np
from numpy.typing import NDArray

from .raycaster import RaycasterCamera, Sprite
from .text import Cell, Point, PosHint, Size, SizeHint, Text, clamp

__all__ = ["TextRaycaster", "RaycasterCamera", "Sprite", "Point", "Size"]


class TextRaycaster(Text):
    r"""
    A raycaster gadget.

    Parameters
    ----------
    caster_map : NDArray[np.ushort]
        The raycaster map.
    camera : RaycasterCamera
        The raycaster camera.
    wall_textures : List[NDArray[np.uint]]
        Textures for walls.
    sprites : list[Sprite] | None, default: None
        A list of sprites.
    sprite_textures : list[NDArray[np.str\_]] | None, default: None
        Textures for sprites.
    max_hops : int, default: 20
        Determines how far rays are cast.
    ascii_map : str, default: " .,:;<+*LtCa4U80dQM@"
        Dark to bright ascii characters for shading.
    default_cell : NDArray[Cell] | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    caster_map : NDArray[np.ushort]
        The raycaster map.
    camera : RaycasterCamera
        The raycaster camera.
    wall_textures : List[NDArray[np.uint]]
        Textures for walls.
    sprites : list[Sprite]
        A list of sprites.
    sprite_textures : list[NDArray[np.str\_]]
        Textures for sprites.
    max_hops : int
        Determines how far rays are cast.
    ascii_map : str
        Dark to bright ascii characters for shading.
    canvas : NDArray[Cell]
        The array of characters for the gadget.
    default_cell : NDArray[Cell]
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    cast_rays()
        Update canvas by casting all rays.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        caster_map: NDArray[np.ushort],
        camera: RaycasterCamera,
        wall_textures: list[NDArray[np.uint]] | None,
        sprites: list[Sprite] | None = None,
        sprite_textures: list[NDArray[np.str_]] | None = None,
        max_hops: int = 20,
        ascii_map: str = " .,:;<+*LtCa4U80dQM@",
        default_cell: NDArray[Cell] | str = " ",
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            default_cell=default_cell,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.caster_map = caster_map
        """The raycaster map."""
        self.camera = camera
        """The raycaster camera."""
        self.wall_textures = wall_textures
        """Textures for walls."""
        self.sprites = sprites
        """A list of sprites."""
        self.sprite_textures = sprite_textures
        """Textures for sprites."""
        self.max_hops = max_hops
        """Determines how far rays are cast."""
        self.ascii_map = np.array(list(ascii_map))
        """Dark to bright ascii characters for shading."""

        self._shades = len(self.ascii_map) - 1
        self._shade_values = np.linspace(-12, 24, self._shades, dtype=int)
        self._side_shade = 2
        self._shade_diff = self._shades - self._side_shade
        # Buffers
        self._pos_int = np.zeros((2,), dtype=int)
        self._pos_frac = np.zeros((2,), dtype=float)

        self.on_size()

    def on_size(self):
        """Resize canvas array and re-make caster buffers."""
        h, w = self._size

        self.canvas = np.full((h, w), self.default_cell)

        # Precalculate angle of rays cast.
        self._ray_angles = angles = np.ones((w, 2), dtype=float)
        angles[:, 1] = np.linspace(-1, 1, w)

        # Precalculate distances for ceiling and floor textures
        self._distances = distances = np.linspace(
            1e-10, h, num=h, endpoint=False, dtype=float
        )
        np.divide(h, distances, out=distances, dtype=float)

        # Buffers
        self._rotated_angles = np.zeros_like(angles)
        self._deltas = np.zeros_like(angles)
        self._sides = np.zeros_like(angles)
        self._steps = np.zeros_like(angles, dtype=int)
        self._weights = weights = np.zeros((h, 2), dtype=float)
        self._tex_frac = np.zeros_like(weights)
        self._tex_frac_2 = np.zeros_like(weights)
        self._tex_int = np.zeros_like(weights, dtype=int)
        self._column_distances = np.zeros((w,), dtype=float)

    def cast_rays(self):
        """Update canvas by casting all rays."""
        h, w = self.size
        if h == 0 or w == 0:
            return

        # Early calculations on rays can be vectorized:
        np.dot(self._ray_angles, self.camera._plane, out=self._rotated_angles)
        with np.errstate(divide="ignore"):
            np.true_divide(1.0, self._rotated_angles, out=self._deltas)
        np.absolute(self._deltas, out=self._deltas)
        np.sign(self._rotated_angles, out=self._steps, casting="unsafe")
        np.heaviside(self._steps, 1.0, out=self._sides)
        np.mod(self.camera.pos, 1.0, out=self._pos_frac)
        np.subtract(self._sides, self._pos_frac, out=self._sides)
        np.multiply(self._sides, self._steps, out=self._sides)
        np.multiply(self._sides, self._deltas, out=self._sides)

        self.canvas["char"] = " "
        self.canvas["char"][self.height // 2 :, ::2] = self.ascii_map[1]

        for column in range(self.width):
            self._cast_ray(column)

        self._cast_sprites()

    def _cast_ray(self, column):
        """Cast a ray for a given column of the screen."""
        camera = self.camera
        camera_pos = camera.pos
        caster_map = self.caster_map

        ray_pos = self._pos_int
        ray_pos[:] = camera_pos

        ray_angle = self._rotated_angles[column]
        delta = self._deltas[column]
        step = self._steps[column]
        sides = self._sides[column]

        # Cast a ray until we hit a wall or hit max_hops
        for _ in range(self.max_hops):
            side = 0 if sides[0] < sides[1] else 1
            sides[side] += delta[side]
            ray_pos[side] += step[side]

            if texture_index := caster_map[tuple(ray_pos)]:
                # Distance from wall to camera plane.
                # Note that distance of wall to camera is not used
                # as it would result in a "fish-eye" effect.
                distance = (
                    ray_pos[side] - camera_pos[side] + (0 if step[side] == 1 else 1)
                ) / ray_angle[side]
                break
        else:  # No walls in range.
            distance = 10000

        self._column_distances[column] = distance

        column_height = int(self.height / distance) if distance else 10000
        if column_height == 0:
            return

        # Start and end y-coordinates of column.
        half_height = self.height >> 1
        half_column = column_height >> 1
        if half_column > half_height:
            half_column = half_height

        start = half_height - half_column
        end = half_height + half_column
        drawn_height = end - start

        shade = min(drawn_height, self._shade_diff)
        if side:
            shade += self._side_shade

        shade_buffer = np.full(drawn_height, shade)
        wall_texture = self.wall_textures[texture_index - 1]
        tex_h, tex_w = wall_texture.shape

        # Exactly where wall was hit by ray as a percentage of its width.
        wall_x = (camera_pos[1 - side] + distance * ray_angle[1 - side]) % 1

        # Use above percentage to grab the column of the texture we need.
        tex_x = int(wall_x * tex_w)
        if (-1 if side == 1 else 1) * ray_angle[side] < 0:  # Sign correction.
            tex_x = tex_w - tex_x - 1

        # Interpolate texture onto column
        offset = (column_height - drawn_height) / 2
        ratio = tex_h / column_height
        texture_start = offset * ratio
        texture_end = (offset + drawn_height) * ratio
        tex_ys = np.linspace(
            texture_start, texture_end, num=drawn_height, endpoint=False, dtype=int
        )
        shade_buffer += self._shade_values[wall_texture[tex_ys, tex_x]]
        np.clip(shade_buffer, 1, self._shades, out=shade_buffer)

        # Paint column.
        self.canvas["char"][:, ::-1][start:end, column] = self.ascii_map[shade_buffer]

    def _cast_sprites(self):
        """Render all sprites."""
        h, w = self.size
        half_w = w / 2

        camera = self.camera
        camera_pos = camera.pos
        sprites = self.sprites
        sprite_textures = self.sprite_textures
        column_distances = self._column_distances

        for sprite in sprites:
            sprite.relative = camera_pos - sprite.pos

        sprites.sort()

        # Camera Inverse used to calculate transformed position of sprites.
        cam_inv = np.linalg.inv(-camera._plane)

        # Draw each sprite from furthest to closest.
        for sprite in sprites:
            # Transformed position of sprites due to camera position.
            y, x = sprite.relative @ cam_inv

            if y <= 0:
                # Sprite is behind camera, don't draw it.
                continue

            # Sprite x-position on screen
            sprite_x = int(half_w * (1 + x / y))
            sprite_height = int(h / y)
            sprite_width = int(half_w / y)
            # Is sprite too small?
            if sprite_height == 0 or sprite_width == 0:
                continue

            start_x = clamp(sprite_x - sprite_width // 2, 0, w)
            end_x = clamp(sprite_x + sprite_width // 2, 0, w)
            columns = np.arange(start_x, end_x)
            columns = columns[y <= column_distances[columns]]

            start_y = clamp(int((h - sprite_height) / 2), 0, h)
            end_y = clamp(int((h + sprite_height) / 2), 0, h)
            rows = np.arange(start_y, end_y, dtype=float)

            sprite_tex = sprite_textures[sprite.texture_idx]
            tex_height, tex_width = sprite_tex.shape

            clip_y = (sprite_height - h) / 2
            rows += clip_y
            rows *= tex_height / sprite_height
            np.clip(rows, 0, None, out=rows)

            clip_x = sprite_x - sprite_width / 2
            tex_xs = columns - clip_x
            tex_xs *= tex_width / sprite_width

            sprite_rect = sprite_tex[rows.astype(int)][:, tex_xs.astype(int)]
            chars = self.canvas["char"][:, ::-1]
            chars[start_y:end_y, columns] = np.where(
                sprite_rect != "0",
                sprite_rect,
                chars[start_y:end_y, columns],
            )



"""Functions for blending colors and creating color gradients."""

from math import sin, tau

from ..geometry import lerp
from .color_types import AColor, Color

__all__ = ["darken_only", "lighten_only", "lerp_colors", "gradient", "rainbow_gradient"]


def darken_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the minimum of each channel in `a` and `b`.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with smallest components of `a` and `b`.
    """
    color = (min(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lighten_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the maximum of each channel in `a` and `b`.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with largest components of `a` and `b`.
    """
    color = (max(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lerp_colors[T: (Color, AColor, tuple[int, ...])](a: T, b: T, p: float) -> T:
    """
    Linear interpolation from `a` to `b` with proportion `p`.

    Parameters
    ----------
    a : (Color, AColor, tuple[int, ...])
        A color.
    b : (Color, AColor, tuple[int, ...])
        A color.
    p : float
        Proportion from a to b.

    Returns
    -------
    (Color, AColor, tuple[int, ...])
        The linear interpolation of `a` and `b`.
    """
    color = (round(lerp(c1, c2, p)) for c1, c2 in zip(a, b))
    if isinstance(a, (Color, AColor)):
        return type(a)(*color)
    return tuple(color)


def gradient[T: (Color, AColor, tuple[int, ...])](
    start: T, end: T, ncolors: int
) -> list[T]:
    """
    Return a gradient from `start` to `end` with `ncolors` (> 1) colors.

    Parameters
    ----------
    start : (Color, AColor, tuple[int, ...])
        Start color of gradient.
    end : (Color, AColor, tuple[int, ...])
        End color of gradient.
    ncolors : int
        Number of colors in gradient.

    Returns
    -------
    list[(Color, AColor, tuple[int, ...])]
        A gradient of colors from `start` to `end`.
    """
    if ncolors < 2:
        raise ValueError(f"not enough colors ({ncolors=})")

    return [lerp_colors(start, end, i / (ncolors - 1)) for i in range(ncolors)]


def rainbow_gradient(n: int, *, alpha: int | None = None) -> list[Color | AColor]:
    """
    Return a rainbow gradient of ``n`` colors.

    Parameters
    ----------
    n : int
        Number of colors in gradient.
    alpha : int | None, default: None
        If ``alpha`` is not given, gradient colors will have no alpha channel.
        Otherwise, the color's alpha channel is given by ``alpha``.

    Returns
    -------
    list[Color | AColor]
        A rainbow gradient of colors.
    """
    theta = tau / n
    offsets = [0, tau / 3, 2 * tau / 3]

    def color(i):
        return (int(sin(i * theta + offset) * 127 + 128) for offset in offsets)

    if alpha is None:
        return [Color(*color(i)) for i in range(n)]

    return [AColor(*color(i), alpha) for i in range(n)]



"""A raycaster gadget."""

from dataclasses import dataclass
from typing import Literal, Protocol

import numpy as np
from numpy.typing import NDArray

from ..colors import BLACK, TRANSPARENT, AColor, Color
from ..texture_tools import _composite
from .graphics import Graphics, Interpolation, Point, PosHint, Size, SizeHint, clamp

__all__ = ["Raycaster", "Sprite", "RaycasterCamera", "RgbaTexture", "Point", "Size"]


@dataclass
class Sprite:
    """A sprite for a raycaster."""

    pos: tuple[float, float]
    """Position of sprite on map."""
    texture_idx: int
    """Index of sprite texture."""

    def __post_init__(self):
        self.relative: NDArray[np.float64] = np.zeros(2)

    @property
    def relative(self) -> NDArray[np.float64]:
        """Vector from camera to sprite."""
        return self._relative

    @relative.setter
    def relative(self, relative):
        self._relative = relative
        self.distance = relative @ relative

    def __lt__(self, other):
        """Sprites are ordered by their distance to camera."""
        return self.distance > other.distance


class RaycasterCamera:
    """
    A raycaster camera.

    Parameters
    ----------
    pos : tuple[float, float], default: (0.0, 0.0)
        Position of camera on the map.
    theta : float, default: 0.0
        Direction of camera in radians.
    fov : float, default: 0.66
        Field of view of camera.

    Attributes
    ----------
    pos : tuple[float, float]
        Position of camera on the map.
    theta : float
        Direction of camera in radians.
    fov : float
        Field of view of camera.

    Methods
    -------
    rotation_matrix(theta)
        Return a 2-D rotation matrix from a given angle.
    rotate(theta)
        Rotate camera `theta` radians in-place.
    """

    def __init__(
        self,
        pos: tuple[float, float] = (0.0, 0.0),
        theta: float = 0.0,
        fov: float = 0.66,
    ):
        self.pos = np.array(pos, float)
        self._build_plane(theta, fov)

    @staticmethod
    def rotation_matrix(theta: float) -> NDArray[np.float32]:
        """Return a 2-D rotation matrix from a given angle."""
        x = np.cos(theta)
        y = np.sin(theta)
        return np.array([[x, y], [-y, x]], float)

    def _build_plane(self, theta: float, fov: float) -> NDArray[np.float64]:
        initial_plane = np.array([[1.001, 0.001], [0.0, fov]], float)
        self._plane = initial_plane @ self.rotation_matrix(theta)

    @property
    def theta(self) -> float:
        """Direction of camera in radians."""
        x2, x1 = self._plane[0]
        return np.arctan2(x1, x2)

    @theta.setter
    def theta(self, theta: float):
        self._build_plane(theta, self.fov)

    @property
    def fov(self) -> float:
        """Field of view of camera."""
        return (self._plane[1] ** 2).sum() ** 0.5

    @fov.setter
    def fov(self, fov: float):
        self._build_plane(self.theta, fov)

    def rotate(self, theta: float):
        """Rotate camera `theta` radians."""
        self._plane = self._plane @ self.rotation_matrix(theta)


class RgbaTexture(Protocol):
    """
    A RGBA texture. Typically a numpy array.

    Notes
    -----
    This protocol is provided to allow for, say, animated textures. The raycaster
    will function as long as `shape` and `__getitem__` work as expected.
    """

    shape: tuple[int, int, Literal[4]]  # (height, width, RGBA)

    def __getitem__(self, key) -> NDArray[np.uint8] | np.uint8:
        """Supports numpy indexing."""


class Raycaster(Graphics):
    r"""
    A raycaster gadget.

    Parameters
    ----------
    caster_map : NDArray[np.ushort]
        The raycaster map.
    camera : RaycasterCamera
        The raycaster camera.
    wall_textures : List[RgbaTexture]
        Textures for walls.
    light_wall_textures : list[RgbaTexture] | None, default: None
        Optional wall textures for north/south facing walls.
    sprites : list[Sprite] | None, default: None
        A list of sprites.
    sprite_textures : list[RgbaTexture] | None, default: None
        Textures for sprites.
    ceiling : RgbaTexture | None, default: None
        Optional ceiling texture.
    ceiling_color : Color, default: BLACK
        Color of ceiling if no ceiling texture.
    floor : RgbaTexture | None, default: None
        Optional floor texture.
    floor_color : Color, default: BLACK
        Color of floor if no floor texture.
    max_hops : int, default: 20
        Determines how far rays are cast.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    caster_map : NDArray[np.ushort]
        The raycaster map.
    camera : RaycasterCamera
        The raycaster camera.
    wall_textures : List[RgbaTexture]
        Textures for walls.
    light_wall_textures : list[RgbaTexture]
        Wall textures for north/sourth facing walls.
    sprites : list[Sprite]
        A list of sprites.
    sprite_textures : list[RgbaTexture]
        Textures for sprites.
    ceiling : RgbaTexture | None
        The ceiling texture.
    ceiling_color : Color
        Color of ceiling if no ceiling texture.
    floor : RgbaTexture
        The floor texture.
    floor_color : Color
        Color of floor if no floor texture.
    max_hops : int
        Determines how far rays are cast.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    cast_rays()
        Update texture by casting all rays.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        caster_map: NDArray[np.ushort],
        camera: RaycasterCamera,
        wall_textures: list[RgbaTexture] | None,
        light_wall_textures: list[RgbaTexture] | None = None,
        sprites: list[Sprite] | None = None,
        sprite_textures: list[RgbaTexture] | None = None,
        ceiling: RgbaTexture | None = None,
        ceiling_color: Color = BLACK,
        floor: RgbaTexture | None = None,
        floor_color: Color = BLACK,
        max_hops: int = 20,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.caster_map = caster_map
        """The raycaster map."""
        self.camera = camera
        """The raycaster camera."""
        self.wall_textures = wall_textures
        """Textures for walls."""
        self.light_wall_textures = light_wall_textures or wall_textures
        """Optional wall textures for north/south facing walls."""
        self.sprites = sprites
        """A list of sprites."""
        self.sprite_textures = sprite_textures
        """Textures for sprites."""
        self.ceiling = ceiling
        """Optional ceiling texture."""
        self.ceiling_color = ceiling_color
        """Color of ceiling if no ceiling texture."""
        self.floor = floor
        """Optional floor texture."""
        self.floor_color = floor_color
        """Color of floor if no floor texture."""
        self.max_hops = max_hops
        """Determines how far rays are cast."""

        # Buffers
        self._pos_int = np.zeros((2,), dtype=int)
        self._pos_frac = np.zeros((2,), dtype=float)
        self._floor_pos = np.zeros((2,), dtype=float)

        self.on_size()

    def on_size(self):
        """Resize texture array and re-make caster buffers."""
        h, w = self._size

        self.texture = np.zeros((2 * h, w, 4), dtype=np.uint8)

        # Precalculate angle of rays cast.
        self._ray_angles = angles = np.ones((w, 2), dtype=float)
        angles[:, 1] = np.linspace(-1, 1, w)

        # Precalculate distances for ceiling and floor textures
        self._distances = distances = np.linspace(
            1e-10, h, num=h, endpoint=False, dtype=float
        )
        np.divide(h, distances, out=distances, dtype=float)

        # Buffers
        self._rotated_angles = np.zeros_like(angles)
        self._deltas = np.zeros_like(angles)
        self._sides = np.zeros_like(angles)
        self._steps = np.zeros_like(angles, dtype=int)
        self._weights = weights = np.zeros((h, 2), dtype=float)
        self._tex_frac = np.zeros_like(weights)
        self._tex_frac_2 = np.zeros_like(weights)
        self._tex_int = np.zeros_like(weights, dtype=int)
        self._column_distances = np.zeros((w,), dtype=float)

    def cast_rays(self):
        """Update texture by casting all rays."""
        h, w = self.size
        if h == 0 or w == 0:
            return

        # Early calculations on rays can be vectorized:
        np.dot(self._ray_angles, self.camera._plane, out=self._rotated_angles)
        with np.errstate(divide="ignore"):
            np.true_divide(1.0, self._rotated_angles, out=self._deltas)
        np.absolute(self._deltas, out=self._deltas)
        np.sign(self._rotated_angles, out=self._steps, casting="unsafe")
        np.heaviside(self._steps, 1.0, out=self._sides)
        np.mod(self.camera.pos, 1.0, out=self._pos_frac)
        np.subtract(self._sides, self._pos_frac, out=self._sides)
        np.multiply(self._sides, self._steps, out=self._sides)
        np.multiply(self._sides, self._deltas, out=self._sides)

        self.texture[:h, :, :3] = self.ceiling_color
        self.texture[h:, :, :3] = self.floor_color
        self.texture[..., 3] = 255

        for column in range(self.width):
            self._cast_ray(column)

        self._cast_sprites()

    def _cast_ray(self, column):
        """Cast a ray for a given column of the screen."""
        camera = self.camera
        camera_pos = camera.pos
        caster_map = self.caster_map

        ray_pos = self._pos_int
        ray_pos[:] = camera_pos

        ray_angle = self._rotated_angles[column]
        delta = self._deltas[column]
        step = self._steps[column]
        sides = self._sides[column]

        # Cast a ray until we hit a wall or hit max_hops
        for _ in range(self.max_hops):
            side = 0 if sides[0] < sides[1] else 1
            sides[side] += delta[side]
            ray_pos[side] += step[side]

            if texture_index := caster_map[tuple(ray_pos)]:
                # Distance from wall to camera plane.
                # Note that distance of wall to camera is not used
                # as it would result in a "fish-eye" effect.
                distance = (
                    ray_pos[side] - camera_pos[side] + (0 if step[side] == 1 else 1)
                ) / ray_angle[side]
                break
        else:  # No walls in range.
            distance = 10000

        self._column_distances[column] = distance

        texture = self.texture[:, ::-1]
        height = texture.shape[0]

        column_height = int(height / distance) if distance else 10000
        if column_height == 0:
            return

        # Start and end y-coordinates of column.
        half_height = height >> 1
        half_column = column_height >> 1
        if half_column > half_height:
            half_column = half_height

        start = half_height - half_column
        end = half_height + half_column

        wall_texture = (self.wall_textures if side else self.light_wall_textures)[
            texture_index - 1
        ]
        tex_h, tex_w, _ = wall_texture.shape

        # Exactly where wall was hit by ray as a percentage of its width.
        wall_x = (camera_pos[1 - side] + distance * ray_angle[1 - side]) % 1

        # Use above percentage to grab the column of the texture we need.
        tex_x = int(wall_x * tex_w)
        if (-1 if side == 1 else 1) * ray_angle[side] < 0:  # Sign correction.
            tex_x = tex_w - tex_x - 1

        # Interpolate texture onto column
        drawn_height = end - start
        offset = (column_height - drawn_height) / 2
        ratio = tex_h / column_height
        texture_start = offset * ratio
        texture_end = (offset + drawn_height) * ratio
        tex_ys = np.linspace(
            texture_start, texture_end, num=drawn_height, endpoint=False, dtype=int
        )
        texture_column = wall_texture[tex_ys, tex_x].astype(float)

        # Darken colors further away.
        texture_column *= np.e ** (-distance * 0.05)
        np.clip(texture_column, 0, 255, out=texture_column, casting="unsafe")

        # Paint column.
        texture[start:end, column] = texture_column

        # Render floor and ceiling.
        ceiling = self.ceiling
        floor = self.floor

        if ceiling is None and floor is None:
            return

        # Buffer views
        floor_pos = self._floor_pos
        weights = self._weights[half_column:]
        tex_frac = self._tex_frac[half_column:]
        tex_frac_2 = self._tex_frac_2[half_column:]
        tex_int = self._tex_int[half_column:]

        # Floor position
        if side == 0:
            facing = float(ray_angle[0] < 0), wall_x
        else:
            facing = wall_x, float(ray_angle[1] < 0)
        np.add(ray_pos, facing, out=floor_pos)

        # Horizontal distances of floor / ceiling
        np.divide(self._distances[half_column:], distance, out=weights[:, 0])
        weights[:, 1] = weights[:, 0]

        # Texture coordinates
        # (weights * floor_pos + (1 - weights) * camera_pos) % 1
        np.multiply(weights, floor_pos, out=tex_frac)
        np.subtract(1.0, weights, out=weights)
        np.multiply(weights, camera_pos, out=tex_frac_2)
        np.add(tex_frac, tex_frac_2, out=tex_frac)
        np.mod(tex_frac, 1.0, out=tex_frac)

        # Paint ceiling
        if ceiling is not None:
            # Note reversed order of texture coordinates from floor
            np.multiply(
                ceiling.shape[:2], tex_frac[::-1], out=tex_int, casting="unsafe"
            )
            texture[:start, column] = ceiling[tex_int[:, 0], tex_int[:, 1]]

        # Paint floor
        if floor is not None:
            np.multiply(floor.shape[:2], tex_frac, out=tex_int, casting="unsafe")
            texture[end:, column] = floor[tex_int[:, 0], tex_int[:, 1]]

    def _cast_sprites(self):
        """Render all sprites."""
        texture = self.texture[:, ::-1]
        h, w, _ = texture.shape
        half_w = w / 2

        camera = self.camera
        camera_pos = camera.pos
        sprites = self.sprites
        sprite_textures = self.sprite_textures
        column_distances = self._column_distances

        for sprite in sprites:
            sprite.relative = camera_pos - sprite.pos

        sprites.sort()

        # Camera Inverse used to calculate transformed position of sprites.
        cam_inv = np.linalg.inv(-camera._plane)

        # Draw each sprite from furthest to closest.
        for sprite in sprites:
            # Transformed position of sprites due to camera position.
            y, x = sprite.relative @ cam_inv

            if y <= 0:
                # Sprite is behind camera, don't draw it.
                continue

            # Sprite x-position on screen
            sprite_x = int(half_w * (1 + x / y))
            sprite_height = int(h / y)
            sprite_width = int(half_w / y)
            # Is sprite too small?
            if sprite_height == 0 or sprite_width == 0:
                continue

            start_x = clamp(sprite_x - sprite_width // 2, 0, w)
            end_x = clamp(sprite_x + sprite_width // 2, 0, w)
            columns = np.arange(start_x, end_x)
            columns = columns[y <= column_distances[columns]]

            start_y = clamp(int((h - sprite_height) / 2), 0, h)
            end_y = clamp(int((h + sprite_height) / 2), 0, h)
            rows = np.arange(start_y, end_y, dtype=float)

            sprite_tex = sprite_textures[sprite.texture_idx]
            tex_height, tex_width, _ = sprite_tex.shape

            clip_y = (sprite_height - h) / 2
            rows += clip_y
            rows *= tex_height / sprite_height
            np.clip(rows, 0, None, out=rows)

            clip_x = sprite_x - sprite_width / 2
            tex_xs = columns - clip_x
            tex_xs *= tex_width / sprite_width

            sprite_rect = sprite_tex[rows.astype(int)][:, tex_xs.astype(int)]
            dst = texture[start_y:end_y, columns, :3]
            _composite(dst, sprite_rect[..., :3], sprite_rect[..., 3, None])
            texture[start_y:end_y, columns, :3] = dst



"""An animation gadget."""

import asyncio
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Self

import numpy as np
from numpy.typing import NDArray

from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, bindable, clamp
from .image import Image, Interpolation

__all__ = ["Animation", "Interpolation", "Point", "Size"]


def _check_frame_durations(
    frames: list[Image], frame_durations: float | Sequence[float]
) -> Sequence[float]:
    """
    Raise `ValueError` if `frames` and `frame_durations` are incompatible,
    else return a sequence of frame durations.
    """
    if not isinstance(frame_durations, Sequence):
        return [frame_durations] * len(frames)

    if len(frame_durations) != len(frames):
        raise ValueError("number of frames doesn't match number of frame durations")

    return frame_durations


class Animation(Gadget):
    r"""
    An animation gadget.

    Parameters
    ----------
    path : Path | None, default: None
        Path to directory of images for frames in the animation (loaded
        in lexographical order of filenames).
    frame_durations : float | Sequence[float], default: 1/12
        Time each frame is displayed. If a sequence is provided, it's length should be
        equal to number of frames.
    loop : bool, default: True
        Whether to restart animation after last frame.
    reverse : bool, default: False
        Whether to play animation in reverse.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    frames : list[Image]
        Frames of the animation.
    frame_durations : list[int | float]
        Time each frame is displayed.
    loop : bool
        Whether to animation is restarted after last frame.
    reverse : bool
        Whether to animation is played in reverse.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    play()
        Play the animation. Returns a task.
    pause()
        Pause the animation
    stop()
        Stop the animation and reset current frame.
    from_textures(textures, ...)
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.
    from_images(images, ...)
        Create an :class:`Animation` from an iterable of :class:`Image`.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path | None = None,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.frames: list[Image]
        """Frames of the animation."""

        if path is not None:
            paths = sorted(path.iterdir(), key=lambda file: file.name)
            self.frames = [
                Image(path=path, size=size, is_transparent=is_transparent)
                for path in paths
            ]
            for frame in self.frames:
                frame.parent = self
        else:
            self.frames = []

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.frame_durations = _check_frame_durations(self.frames, frame_durations)
        self.alpha = alpha
        self.interpolation = interpolation
        self.loop = loop
        self.reverse = reverse
        self._i = len(self.frames) - 1 if self.reverse else 0
        self._animation_task = None

    def on_remove(self):
        """Pause animation."""
        self.pause()
        super().on_remove()

    def on_size(self) -> None:
        """Update size of all frames on resize."""
        for frame in self.frames:
            frame.size = self._size

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        for frame in self.frames:
            frame.is_transparent = self.is_transparent

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)
        for frame in self.frames:
            frame.alpha = alpha

    @property
    def interpolation(self) -> Interpolation:
        """Interpolation used when gadget is resized."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, interpolation: Interpolation):
        self._interpolation = interpolation
        for frame in self.frames:
            frame.interpolation = interpolation

    async def _play_animation(self):
        while self.frames:
            try:
                await asyncio.sleep(self.frame_durations[self._i])
            except asyncio.CancelledError:
                return

            if self.reverse:
                self._i -= 1
                if self._i < 0:
                    self._i = len(self.frames) - 1

                    if not self.loop:
                        return
            else:
                self._i += 1
                if self._i == len(self.frames):
                    self._i = 0

                    if not self.loop:
                        return

    def play(self) -> asyncio.Task:
        """
        Play animation.

        Returns
        -------
        asyncio.Task
            The task that plays the animation.
        """
        self.pause()

        if self._i == 0 and self.reverse:
            self._i = len(self.frames) - 1
        elif self._i == len(self.frames) - 1 and not self.reverse:
            self._i = 0

        self._animation_task = asyncio.create_task(self._play_animation())
        return self._animation_task

    def pause(self):
        """Pause animation."""
        if self._animation_task is not None:
            self._animation_task.cancel()

    def stop(self):
        """Stop the animation and reset current frame."""
        self.pause()
        self._i = len(self.frames) - 1 if self.reverse else 0

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        if self.frames:
            self.frames[self._i]._region = self._region
            self.frames[self._i]._render(canvas)
        else:
            super()._render(canvas)

    @classmethod
    def from_textures(
        cls,
        textures: Iterable[NDArray[np.uint8]],
        *,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.

        Parameters
        ----------
        textures : Iterable[NDArray[np.uint8]]
            An iterable of RGBA textures that will be the frames of the animation.
        frame_durations : float | Sequence[float], default: 1/12
            Time each frame is displayed. If a sequence is provided, it's length should
            be equal to number of frames.
        loop : bool, default: True
            Whether to restart animation after last frame.
        reverse : bool, default: False
            Whether to play animation in reverse.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Animation
            A new animation gadget.
        """
        animation = cls(
            loop=loop,
            reverse=reverse,
            alpha=alpha,
            interpolation=interpolation,
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        animation.frames = [
            Image.from_texture(
                texture,
                size=animation.size,
                alpha=animation.alpha,
                interpolation=animation.interpolation,
            )
            for texture in textures
        ]
        for frame in animation.frames:
            frame.parent = animation
        animation.frame_durations = _check_frame_durations(
            animation.frames, frame_durations
        )
        return animation

    @classmethod
    def from_images(
        cls,
        images: Iterable[Image],
        *,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Animation` from an iterable of :class:`Image`.

        Parameters
        ----------
        images : Iterable[Image]
            An iterable of images that will be the frames of the animation.
        frame_durations : float | Sequence[float], default: 1/12
            Time each frame is displayed. If a sequence is provided, it's length should
            be equal to number of frames.
        loop : bool, default: True
            Whether to restart animation after last frame.
        reverse : bool, default: False
            Whether to play animation in reverse.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Animation
            A new animation gadget.
        """
        animation = cls(
            loop=loop,
            reverse=reverse,
            alpha=alpha,
            interpolation=interpolation,
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        animation.frames = list(images)
        for image in animation.frames:
            image.interpolation = animation.interpolation
            image.size = animation.size
            image.alpha = animation.alpha
            image.parent = animation
        animation.frame_durations = _check_frame_durations(
            animation.frames, frame_durations
        )
        return animation



"""A shadow caster gadget."""

from bisect import bisect
from collections.abc import Callable
from dataclasses import dataclass, field
from itertools import product
from math import dist
from typing import Literal

import numpy as np
from numpy.typing import NDArray

from ..colors import AWHITE, BLACK, TRANSPARENT, WHITE, AColor, Color
from ..geometry import Region, rect_slice
from ..texture_tools import resize_texture
from .graphics import Graphics, Interpolation, Point, PosHint, Size, SizeHint, clamp

__all__ = [
    "ShadowCaster",
    "Restrictiveness",
    "ShadowCasterCamera",
    "LightSource",
    "Point",
    "Size",
]

AGRAY = AColor(50, 50, 50)
QUADS = tuple(product((1, -1), (1, -1), (False, True)))

Restrictiveness = Literal["permissive", "moderate", "reestrictive"]
"""
The restrictiveness of the shadow caster.

:class:`Restrictiveness` is one of "permissive", "moderate", "restrictive".

For "permissive", any interval is visible as long as any of it's start, center, or end
points are visible. For "moderate", the center and either end must be visible. For
"restrictive", all points in the interval must be visible.
"""


@dataclass(slots=True)
class ShadowCasterCamera:
    """
    The camera determines the visible portion of the map.

    If the camera's size is not equal to the caster's size, the camera will resize the
    visible portion of the map to fill the caster's size. To prevent resizing the
    visible portion of the map, the camera's size should be equal to the caster's size.

    Parameters
    ----------
    pos : Point
        The position of the upper-left corner of the camera on the map.
    size : Size
        Size of the camera.

    Methods
    -------
    get_submap(map)
        Get the section of the map visible by the camera.

    Notes
    -----
    Submap values for areas of the camera that are out-of-bounds of the map will be
    zero.
    """

    pos: Point
    size: Size

    def get_submap(self, map: NDArray[np.uint32]) -> NDArray[np.uint32]:
        """
        Get the section of a map visible by the camera.

        Returns
        -------
        NDArray[np.uint32]
            Section of map visible by the camera.
        """
        submap = np.zeros(self.size, dtype=np.uint8)

        source = Region.from_rect((0, 0), map.shape)
        dest = Region.from_rect(self.pos, self.size)

        if intersection := source & dest:
            pos, size = next(intersection.rects())
            src = rect_slice(pos, size)
            dst = rect_slice(pos - self.pos, size)
            submap[dst] = map[src]

        return submap


@dataclass(slots=True)
class LightSource:
    """
    A light source.

    Parameters
    ----------
    coords : float[tuple, tuple], default: (0.0, 0.0)
        Coordinates of light source on map.
    color : Color, default: WHITE
        Color of light source.

    Attributes
    ----------
    coords : float[tuple, tuple]
        Coordinates of light source on map.
    color : Color
        Color of light source.
    """

    coords: tuple[float, float] = 0.0, 0.0
    color: Color = WHITE


@dataclass(slots=True)
class _Interval:
    """
    A continuous interval.

    Parameters
    ----------
    start : float
        Start of interval.
    end : float
        End of interval.

    Attributes
    ----------
    start : float
        Start of interval.
    end : float
        End of interval.
    """

    start: float
    end: float
    center: float = field(init=False)

    def __post_init__(self):
        self.center = (self.start + self.end) / 2

    def __contains__(self, item: float) -> bool:
        return self.start <= item <= self.end

    def __gt__(self, other):
        if isinstance(other, (float, int)):
            return other < self.start

        return (other.start, other.end) < (self.start, self.end)


class ShadowCaster(Graphics):
    r"""
    A restrictive precise angle shadowcaster.

    :meth:`cast_shadows` must be called to generate or update :attr:`texture`.

    Parameters
    ----------
    map : NDArray[np.uint32]
        A 2-d map. Non-zero values are walls.
    camera : ShadowCasterCamera
        A camera that determines the visible portion of the map.
    tile_colors : list[AColor] | None, default: None
        A value `n` in the map will be colored ``tile_colors[n]``. If ``None``,
        `tile_colors` will be set to ``[AGRAY, AWHITE]``.
    light_sources : list[LightSource] | None, default: None
        A list of all light sources on the map.
    ambient_light : Color, default: BLACK
        Color of ambient light. Default is no light.
    restrictiveness : Restrictiveness, default: "moderate"
        Restrictiveness of casting algorithm.
    light_decay : Callable[[float], float], default: lambda d: 1 if d == 0 else 1 / d
        The strength of light as a function of distance from origin.
    radius : int, default: 20
        Max visible radius from a light source.
    smoothing : float, default: 1.0 / 3.0
        Smoothness of shadow edges. This value will be clamped between `0` and `1`.
    not_visible_blocks : bool, default: True
        Whether all not-visible cells will be treated as opaque.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    map : NDArray[np.uint32]
        A 2-d map. Non-zero values are walls.
    camera : ShadowCasterCamera
        A camera that determines the visible portion of the map.
    tile_colors : list[AColor]
        A value `n` in the map will be colored ``tile_colors[n]``.
    light_sources : list[LightSource]
        A list of all light sources on the map.
    ambient_light : Color
        Color of ambient light.
    restrictiveness : Restrictiveness
        Restrictiveness of casting algorithm.
    light_decay : Callable[[float], float]
        The strength of light as a function of distance from origin.
    radius : int
        Max visible radius from a light source.
    smoothing : float
        Smoothness of shadow edges.
    not_visible_blocks : bool
        Whether all not-visible cells will be treated as opaque.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    cast_shadows()
        Update texture by shadow casting all light sources.
    to_map_coords(point)
        Convert a point in the gadget's local coordinates to a point in the map's
        coordinates.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        map: NDArray[np.uint32],
        camera: ShadowCasterCamera,
        tile_colors: list[AColor] | None = None,
        light_sources: list[LightSource] | None = None,
        ambient_light: Color = BLACK,
        restrictiveness: Restrictiveness = "moderate",
        light_decay: Callable[[float], float] = lambda d: 1 if d == 0 else 1 / d,
        radius: int = 20,
        smoothing: float = 1.0 / 3.0,
        not_visible_blocks: bool = True,
        is_transparent: bool = True,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            is_transparent=is_transparent,
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.map = map
        self.camera = camera
        self.tile_colors = np.array(tile_colors or [AGRAY, AWHITE], dtype=np.uint8)
        self.light_sources = light_sources or []
        self.ambient_light = ambient_light
        self.restrictiveness = restrictiveness
        self.light_decay = light_decay
        self.radius = radius
        self.smoothing = clamp(smoothing, 0.0, 1.0)
        self.not_visible_blocks = not_visible_blocks

    def cast_shadows(self):
        """Update texture by shadow casting all light sources."""
        h, w, _ = self.texture.shape
        if h == 0 or w == 0:
            return

        cy, cx = self.camera.pos
        ch, cw = self.camera.size

        v_scale = h / ch
        h_scale = w / cw

        map = resize_texture(self.camera.get_submap(self.map), (h, w), "nearest")

        # Combined light colors for all light_sources:
        total_light = np.full((h, w, 3), self.ambient_light, dtype=float)

        for light_source in self.light_sources:
            colors = np.zeros_like(total_light)

            # Calculate light position in camera's coordinates:
            ly, lx = light_source.coords
            origin = round(v_scale * (ly - cy)), round(h_scale * (lx - cx))

            color = light_source.color

            for quad in QUADS:
                self._visible_points_quad(quad, origin, color, colors, map)

            total_light += colors

        colored_map = self.tile_colors[map]
        self.texture[..., :3] = colored_map[..., :3] * np.clip(
            total_light / 255, 0.0, 1.0
        )
        self.texture[..., 3] = colored_map[..., 3]

    def _visible_points_quad(self, quad, origin, color, colors, map):
        y, x, vert = quad
        oy, ox = origin
        h, w, _ = colors.shape

        light_decay = self.light_decay
        smooth_radius = self.radius + self.smoothing

        obstructions = []
        for i in range(self.radius):
            if (
                len(obstructions) == 1
                and obstructions[0].start == 0.0
                and obstructions[0].end == 1.0
            ):
                return

            theta = 1.0 / float(i + 1)
            if vert:
                py = oy + i * y
            else:
                px = ox + i * x

            for j in range(i + 1):
                if vert:
                    px = ox + j * x
                else:
                    py = oy + j * y

                p = py, px

                if not (0 <= py < h and 0 <= px < w):
                    continue

                if (d := dist(origin, p)) <= smooth_radius:
                    interval = _Interval(j * theta, (j + 1) * theta)

                    if self._point_is_visible(interval, obstructions):
                        colors[p] = color
                        colors[p] *= light_decay(d)

                        if map[p] != 0:
                            self._add_obstruction(obstructions, interval)

                    elif self.not_visible_blocks:
                        self._add_obstruction(obstructions, interval)

    def _point_is_visible(self, interval: _Interval, obstructions):
        start_visible = center_visible = end_visible = True

        a = bisect(obstructions, interval.start)
        if a > 0:
            a -= 1

        b = bisect(obstructions, interval.end)
        if b < len(obstructions):
            b += 1

        for i in range(a, b):
            obstruction = obstructions[i]

            if start_visible and interval.start in obstruction:
                start_visible = False

            if center_visible and interval.center in obstruction:
                center_visible = False

            if end_visible and interval.end in obstruction:
                end_visible = False

        if self.restrictiveness == "permissive":
            return center_visible or start_visible or end_visible
        if self.restrictiveness == "moderate":
            return center_visible and (start_visible or end_visible)
        if self.restrictiveness == "restrictive":
            return center_visible and start_visible and end_visible

    def _add_obstruction(self, obstructions, obstruction: _Interval):
        start = obstruction.start
        end = obstruction.end

        a = bisect(obstructions, start)
        b = bisect(obstructions, end)

        if a > 0 and start <= obstructions[a - 1].end:
            start = obstructions[a - 1].start
            a -= 1

        if b < len(obstructions) and obstructions[b].end <= end:
            end = obstructions[b].end
            b += 1
        elif b > 0 and end < obstructions[b - 1].end:
            end = obstructions[b - 1].end

        if a == b:
            obstructions.insert(a, _Interval(start, end))
        else:
            obstructions[a:b] = [_Interval(start, end)]

    def to_map_coords(self, point: Point) -> tuple[float, float]:
        """
        Convert a point in the gadget's local coordinates to a point in the map's
        coordinates.

        Parameters
        ----------
        point : Point
            Point in local coordinates.

        Returns
        -------
        tuple[float, float]
            The coordinates of the point in the map.
        """
        y, x = point
        h, w = self.size

        cy, cx = self.camera.pos
        ch, cw = self.camera.size

        return ch / h * y + cy, cw / w * x + cx
